#!/usr/bin/perl -w

# Copyright 2013, 2014, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Print a tikz picture which is all UArD tree row lines up to an X,Y limit.
# The row lines are all overlaid on each other.
#
# This is implemented by going through all $x,$y points within the desired
# X,Y region and checking with Math::PlanePath::PythagoreanTree whether
# they're a visited point.  The line segment N-1 to N and N to N+1 are then
# accumulated in a Set::IntSpan::Fast of covered runs.  Those runs are then
# emited as line segments.
#
# The proof in the paper means that the lines actually have a simple
# structure which could be generated much more efficiently, but this code
# does it with actual tree calculations.


use 5.010;
use strict;
use FindBin;
use List::Util 'max','min';

use Math::PlanePath::PythagoreanTree;
use Math::BigInt try => 'GMP';
use Set::IntSpan::Fast;

use FindBin;
use lib $FindBin::Bin;
use MyPrintwrap ':all';
use MyTriples 'line_is_horizontal';

# uncomment this to run the ### lines
# use Smart::Comments;


sub xy_unshear {
  my ($x,$y) = @_;
  return ($x-$y, $y);
}


my $x_limit = 59;
my $y_limit = 20;

my $path = Math::PlanePath::PythagoreanTree->new (tree_type => 'UArD',
                                                  coordinates => 'PQ');
my $n_start = $path->n_start;

print "  % generated by tools/$FindBin::Script\n";
{
  print "  % horizontals\n";
  $Printwrap = 2;
  printwrap("\\draw");
  $Printwrap += 2;
  my $want_newline;

  foreach my $y1 (0 .. $y_limit) {
    my $set = Set::IntSpan::Fast->new;
    foreach my $x1 (0 .. $x_limit) {
      my $n = $path->xy_to_n(Math::BigInt->new($x1),Math::BigInt->new($y1))
        // next;
      next if $n == $n_start;
      my $depth = $path->tree_n_to_depth($n);
      ### point: "$x1,$y1  depth=$depth"

      foreach my $to_n ($n+1, $n-1) {
        my $to_depth = $path->tree_n_to_depth($to_n);
        next unless $depth == $to_depth;

        my ($x2,$y2) = $path->n_to_xy($to_n);
        next unless line_is_horizontal($x1,$y1, $x2,$y2);
        $x2 = $x2->numify;
        $y2 = $y2->numify;

        if ($x1 > $x2) { ($x1,$x2) = ($x2,$x1); }
        if ($x2 > $x_limit) { $x2 = $x_limit+1; }
        $x2--;
        $set->add_range("$x1", "$x2");
      }
    }

    my @strs;
    my $iter = $set->iterate_runs();
    while (my ($x1, $x2) = $iter->()) {
      $x2++;
      my $y2 = $y1;

      if ($x2 > $x_limit) {
        $x2 = $x_limit;
      }
      next unless $x1 < $x2;
      {
        # my ($x1,$y1) = xy_unshear($x1,$y1);
        # my ($x2,$y2) = xy_unshear($x2,$y2);
        push @strs, " ($x1,$y1)--($x2,$y2)";
      }
    }
    if (@strs) {
      if ($want_newline) {
        printwrap_newline();
      }
      $want_newline = scalar(@strs);
      printwrap (@strs);
    }

    # if (! $set->is_empty) {
    #   printwrap_indent("  % ");
    #   printwrap("y=$y1 lengths");
    #   my $iter = $set->iterate_runs();
    #   while (my ($x1, $x2) = $iter->()) {
    #     $x2++;
    #     my $len = $x2 - $x1 + 1;
    #     printwrap(" $len,");
    #   }
    #   printwrap_newline();
    #   printwrap_indent(" ");
    # }
  }
  printwrap(";");
}
{
  printwrap_newline();
  print "  % diagonals\n";
  $Printwrap = 2;
  printwrap("\\draw");
  $Printwrap += 2;
  my $want_newline;

  my $d_limit = max($x_limit,$y_limit);
  foreach my $d (0 .. $d_limit) {
    my $set = Set::IntSpan::Fast->new;
    foreach my $y1 (0 .. $y_limit) {
      my $x1 = $y1 + $d;
      my $n = $path->xy_to_n(Math::BigInt->new($x1),Math::BigInt->new($y1))
        // next;
      next if $n == $n_start;
      my $depth = $path->tree_n_to_depth($n);

      foreach my $to_n ($n+1, $n-1) {
        my $to_depth = $path->tree_n_to_depth($to_n);
        next unless $depth == $to_depth;

        my ($x2,$y2) = $path->n_to_xy($to_n);
        $x2 = $x2->numify;
        $y2 = $y2->numify;
        next unless line_is_diagonal($x1,$y1, $x2,$y2);

        ($x1,$y1, $x2,$y2) = line_sort_ends($x1,$y1, $x2,$y2);
        if ($y2 > $y_limit) { $y2 = $y_limit+1; }
        $set->add_range($y1, $y2-1);
      }
    }

    my @strs;
    my $iter = $set->iterate_runs();
    while (my ($y1, $y2) = $iter->()) {
      $y2++;
      my $x1 = $y1 + $d;      # d=X-Y so X=d+Y
      my $x2 = $y2 + $d;

      if ($x2 > $x_limit) {
        my $s = $x2 - $x_limit;
        $x2 -= $s;
        $y2 -= $s;
      }
      if ($y2 > $y_limit) {
        my $s = $y2 - $y_limit;
        $x2 -= $s;
        $y2 -= $s;
      }
      next unless $x1 < $x2;
      {
        # my ($x1,$y1) = xy_unshear($x1,$y1);
        # my ($x2,$y2) = xy_unshear($x2,$y2);
        push @strs, " ($x1,$y1)--($x2,$y2)";
      }
    }
    if (@strs) {
      if ($want_newline) {
        printwrap_newline();
      }
      $want_newline = scalar(@strs);
      printwrap (@strs);
    }

    # if (! $set->is_empty) {
    #   printwrap_indent("  % ");
    #   printwrap("d=$d lengths");
    #   my $iter = $set->iterate_runs();
    #   while (my ($y1, $y2) = $iter->()) {
    #     $y2++;
    #     my $len = $y2 - $y1 + 1;
    #     printwrap(" $len,");
    #   }
    #   printwrap_newline();
    #   printwrap_indent(" ");
    # }
  }
  printwrap(";\n");
}
printwrap_newline();
print "  % end generated\n";

exit 0;
