#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2018, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.


# Usage: perl my-check-tex.pl filename...
#
# Check for some mistakes in the given TeX or LaTeX files.

use 5.010;
use strict;
use warnings;
use File::Basename;
use File::Slurp;
use Getopt::Long;
use List::Util 'max','min','sum';
use Math::NumSeq::OEIS::File;
use Regexp::Common 'balanced';
use Tie::IxHash;

use FindBin;
use lib::abs $FindBin::Bin;

# uncomment this to run the ### lines
# use Smart::Comments;

our $VERSION = 0;
$|=1;

my $progname = $FindBin::Script;
my $verbose = 0;

# ENHANCE-ME: \% would be a literal percent, if odd num backslashes
my $comment_re = qr/%[^\n]*\n/;         # TeX comment % through to newline
my $comments_re = qr/($comment_re)*/o;  # zero or more TeX comments

# TeX Book chapter 24 page 270 dimensions syntax spec
my $internal_unit_re = qr/e[mx]/;
my $physical_unit_re = qr/bp|c[cm]|dd|p[ct]|in|m[mu]|sp/;
my $unit_re = qr/$internal_unit_re|(true )?($physical_unit_re)/;
my $number_re = qr/[+-]?[ \t]*[0-9]*([0-9]|[.,])[0-9]*/;
my $dimension_re = qr/$number_re[ \t]*$unit_re/o;
my $kern_sans_backslash_re = qr/m?kern[ \t]*$dimension_re/o;
my $kern_re = qr/\\$kern_sans_backslash_re/o;
my $hspace_re = qr/\\hspace\{[^}]*\}/;
{
  "em" =~ /^$unit_re$/ or die;
  ".1" =~ /^$number_re$/ or die;
  "1.23" =~ $number_re or die;
  "mu" =~ $unit_re or die;
  "1mu" =~ $dimension_re or die;
  ".1em" =~ /^$dimension_re$/ or die;
  "mkern1mu" =~ $kern_sans_backslash_re or die;
  "\\mkern1mu" =~ $kern_re or die;
}

my $braces_re = $RE{'balanced'}{-parens=>'{}'};

my $tex_filename;
my $bad = 0;
sub bad {
  my $linenum = shift;
  $bad = 1;
  print "$tex_filename:$linenum: ",@_,"\n";
}
sub bad_but_not_fatal {
  my $linenum = shift;
  print "$tex_filename:$linenum: ",@_,"\n";
}

sub tex_strip_comments {
  my ($str) = @_;
  # even number of backslashes  \\\\% ...
  # delete the % onwards, keep the backslashes
  $str =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
  return $str;
}
CHECK {
  tex_strip_comments('% foo') eq '' or die;
  tex_strip_comments("% foo\nbar") eq "\nbar" or die;
  tex_strip_comments("a% foo\nbar") eq "a\nbar" or die;
  tex_strip_comments("a\\% foo") eq "a\\% foo" or die;
  tex_strip_comments("a\\\\% foo") eq "a\\\\" or die;
  tex_strip_comments("a\\\\\\% foo") eq "a\\\\\\% foo" or die;
}

sub check_file {
  my $tex_str = File::Slurp::slurp($tex_filename);
  
  # comment is %, literal percent is \%
  # delete when preceded by even num backslashes
  my $tex_str_sans_comments = $tex_str;
  $tex_str_sans_comments =~ s/((^|[^\\])(\\\\)*)%[^\n]*/$1/mg;
  
  # count_lines($tex_str) == count_lines($tex_str_sans_comments)
  #   or die "oops, line count changed";
  
  my $tex_str_sans_comments_and_tikz = $tex_str_sans_comments;
  my $count_tikz = 0;
  my $count_tikz_lines = 0;
  $tex_str_sans_comments_and_tikz
    =~ s{\\begin\{tikzpicture\}.*?\\end\{tikzpicture\}|\\tikzmath$braces_re}
        { my $str = $&;
          $count_tikz++;
          $count_tikz_lines += count_lines($str);
          str_leave_newlines($str)
        }ges;
  
  if ($verbose) {
    print "$tex_filename ",count_lines($tex_str)," lines, $count_tikz tikz of $count_tikz_lines lines\n";
  }
  
  # all \hyphenation{foo-bar qu-ux}, with hyphens deleted for the keys
  my %hyphenation;
  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\hyphenation\{(.*?)\}/g) {
      my $linenum = pos_linenum($str, $-[0]);
      my $h = $1;
      $h =~ s/^\s+//;
      $h =~ s/\s$//;
      foreach my $word (split /\s+/, $h) {
        $word =~ s/-//g;
        $hyphenation{$word} = $linenum;
      }
    }
  }
  
  #----------------------------------------------------------------------------
  # check for % GP-DEFINE redefinitions
  
  {
    my $gp_comment_re = qr{\\\\|/\*};
    my $line_to_define = sub {
      my ($gp) = @_;
      ### line_to_define(): $gp
      if ($gp =~ /([A-Za-z0-9_]+)\s*=\s*memoize\(\1\s*\)/) {
        ### foo=memoize(foo) ...
        return undef;
      }
      if ($gp =~ m{^\s*$gp_comment_re}) {
        ### comment ...
        return undef;
      }
      while ($gp =~ /(?<builtin> for|forstep|my)\(
                   |(?<define>  [a-z0-9_]+)(\(.*\))?\s*=/ixg) {
        ### match ...
        ### builtin: $+{'builtin'}
        ### define : $+{'define'}
        last if defined $+{'builtin'};
        return $+{'define'};
      }
      return undef;
    };
    my %redefinition_ok = ();
    my $str = $tex_str;
    my $continuation = 0;
    my $in_brace = 0;
    my $define = '';   # name of the current function definition
    my $count_defines = 0;
    my %seen;
    while ($str =~ /^% (?:Test-Pari|GP)-DEFINE(.*)/img) {
      my $gp = $1;
      $gp =~ s/^\s+//;  # delete leading whitespace
      ### gp line: $gp
      ### $continuation
      ### $in_brace
      unless ($continuation || $in_brace) {
        if (defined (my $this = $line_to_define->($gp))) {
          $count_defines++;
          $define = $this;
          ### define of: $define
          if (! $redefinition_ok{$define} && $seen{$define}++) {
            my $linenum = pos_linenum($str);
            bad_but_not_fatal($linenum, "GP-DEFINE redefinition: $define");
          }
        }
      }
      $continuation = ($gp =~ /\\\s*$/);
      if ($gp =~ /\{/) { $in_brace = 1; }
      if ($gp =~ /\}/) { $in_brace = 0; }
      
      # error("foo() is blah blah ...")
      if ($gp =~ /error\(\"(oops, )?(.*)\(\)/) {
        my $error_name = $2;
        ### error message from : $error_name
        if ($error_name ne $define) {
          my $linenum = pos_linenum($str);
          bad($linenum, "error() message about \"$error_name\" but in function \"$define\"");
        }
      }
    }
    print "$count_defines gp-inline functions and variables\n";
  }
  
  
  #----------------------------------------------------------------------------
  # OEIS Axxxxxx_samples without an OFFSET recorded
  
  # Not sure always care about this ...
  # {
  #   my $str = $tex_str;
  #   while ($str =~ /\\\n% GP-Test\s+(A\d{6,7})_samples(.*OFFSET=)?/g) {
  #     my $anum = $1;
  #     my $offset = $2;
  #     unless (defined $offset) {
  #       my $linenum = pos_linenum($str, $+[0]);
  #       bad($linenum, "${anum}_samples without intended OFFSET recorded");
  #     }
  #   }
  # }
  
  
  #----------------------------------------------------------------------------
  # OEIS Axxxxxx_samples OFFSET claimed
  
  {
    my $str = $tex_str;
    while ($str =~ /(A\d{6,7})_samples.*OFFSET\s*=\s*(-?\d+)/g) {
      my $anum = $1;
      my $offset = $2;
      ### match: "$anum OFFSET=$offset"
      my $linenum = pos_linenum($str, $+[0]);
      my $seq = Math::NumSeq::OEIS::File->new (anum => $anum);
      unless ($seq->i_start == $offset) {
        bad($linenum, "$anum shown OFFSET=$offset but sequence is actually ",$seq->i_start);
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # tikzpicture various
  # no bare node or coordinate after ;
  #
  # No need for ; after \foreach like \foreach \x in { ... }; <---
  # This shows up as "Missing character: There is no ; in font nullfont!"
  # in the .log file (like other unrecognised stuff in tikzpicture).
  #
  # Do need ; at the end of \path foreach ... {}; the same as other \path
  # of course.
  
  {
    my %exclusive_keys = (left => 1,
                          right => 1,
                          above => 1,
                          below => 1,
                          'above right' => 1,
                          'above left' => 1,
                          'below right' => 1,
                          'below left' => 1,
                          anchor => 1);
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\begin\{tikzpicture\}(.*?)\\end\{tikzpicture\}/sg) {
      my $str_linenum = pos_linenum($str, $-[1]) - 1;
      my $tikz = $1;
      
      # excluding ;; here, which is picked up above separately everywhere
      while ($tikz =~ /;\s*([^[:space:]\\}&;]\w*)|$braces_re/sg) {
        if (defined (my $word = $1)) {
          my $linenum = $str_linenum + pos_linenum($tikz);
          bad($linenum, "tikzpicture ; followed by non-macro: $word");
        } else {
          # print "skip $&\n";
        }
      }
      
      while ($tikz =~ /\\foreach\s+\\\w+\s+in[^{]*?($braces_re\s*)+;/sg) {
        my $linenum = $str_linenum + pos_linenum($tikz);
        bad($linenum, "tikzpicture no need for ; after \\foreach");
      }
      
      # no multiple right,left,anchor etc
      while ($tikz =~ /\[(.*?)\]/sg) {
        my $args = $1;
        my $linenum = $str_linenum + pos_linenum($tikz);
        my $seen;
        ### $args
        foreach my $key (split /,\s*/, $args) {
          # ENHANCE-ME: Allow arg value balanced parens like actual pgf parse.
          $key =~ s/=.*//s;
          ### $key
          next unless $exclusive_keys{$key};
          if ($seen) {
            bad($linenum, "tikzpicture both $seen and $key is unclear");
          } else {
            $seen = $key;
          }
        }
      }
      
      {
        # align=left etc without \\ using that
        my $tikz_trimmed = $tikz;
        $tikz_trimmed =~ s{\.style=$braces_re}{str_leave_newlines($&)}ges;
        while ($tikz_trimmed =~ /\balign=(.*?)(?<!\\);/sg) {
          my $rest = $1;
          next if defined $2;
          my $linenum = $str_linenum + pos_linenum($tikz_trimmed, $-[0]);
          unless ($rest =~ /\\\\|#\d/) {
            bad($linenum, "align= without \\\\ using it");
          }
        }
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # mathtools showonlyrefs=true and showonlyrefs=false should alternate,
  # not repeat
  
  {
    my $str = $tex_str_sans_comments;
    my $prev;
    my $prev_linenum;
    while ($str =~ /showonlyrefs=(true|false)/g) {
      my $this = $1;
      my $linenum = pos_linenum($str, $+[0]);
      if (defined $prev && $this eq $prev) {
        bad($linenum, "showonlyrefs is already $this");
        bad($prev_linenum, "  ... previous setting here");
      } else {
        $prev = $this;
        $prev_linenum = $linenum;
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # align etc no \tag and \notag together
  # showonlyrefs=false want always either \label or \notag
  
  {
    my $str = $tex_str_sans_comments;
    
    # strip \begin{cases} ... \end{cases} etc
    $str =~ s{\\begin\{(cases|aligned|pmatrix|matrix\*?|tikzpicture)\}.*?\\end\{\1\}}
             {str_leave_newlines($&)}ges;
    
    my $within = 'end';
    my %seen;
    my $showonlyrefs = 'false';
    my $showonlyrefs_location = 'default';
    while ($str =~ /(?<separator>\\\\
                    |\\intertext
                    |\\(?<within>begin|end)\{(align|equation|gather)\})
                  |(?<showonlyrefs>showonlyrefs(=(?<showonlyrefsvalue>true|false))?)
                  |\\(?<got>tag\*?|notag|label)
                   /xg) {
      if (defined $+{'separator'}) {
        ### separator: $+{'separator'}
        if ($within eq 'begin') {
          my $total = sum(0, values %seen);
          if ($showonlyrefs eq 'false' && $total != 1) {
            my $linenum = pos_linenum($str, $+[0]);
            bad($linenum, "have $total but want exactly 1 of \\label, \\tag, \\notag in showonlyrefs=$showonlyrefs (from $showonlyrefs_location)");
          } elsif ($total > 1) {
            my $linenum = pos_linenum($str, $+[0]);
            bad($linenum, "duplicates: ".join(',', sort keys %seen));
          }
        }
        if (defined $+{'within'}) {
          $within = $+{'within'};
        }
        %seen = ();
      } elsif (defined $+{'showonlyrefs'}) {
        $showonlyrefs = $+{'showonlyrefsvalue'} // 'true';
        $showonlyrefs_location = "line " . pos_linenum($str, $+[0]);
        ### $showonlyrefs
      } else {
        my $got = $+{'got'};
        $got =~ s/\*$//;
        $seen{$got}++;
        ### got: $+{'got'}
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # \begin{theorem}\label{} should be label name "theorem-foo"
  # \MyFigureText\label{} should be label name "figure-foo"
  # Helps catch \autoref vs \eqref mistakes.
  
  {
    my $str = $tex_str;
    # {theorem}\label{foo}
    while ($str =~ /\{theorem\}\\label\{(?<label>[^}]*)\}/g) {
      my $label = $+{'label'};
      unless ($label =~ /^theorem-/) {
        my $linenum = pos_linenum($str, $+[0]);
        bad($linenum, "theorem label should be \"theorem-...\"");
      }
    }
    # \MyFigureText\label{foo}
    # \caption{text}\label{foo}
    while ($str =~ /(\\MyFigureText|\\caption\{[^}]*\}\s*)
                    \\label\{(?<label>[^}]*)\}/sgx) {
      my $label = $+{'label'};
      unless ($label =~ /^figure-/) {
        my $linenum = pos_linenum($str, $+[0]);
        bad($linenum, "figure label name should be \"figure-...\"");
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # prefer \\[] space measures to have blank line before (blank comment line)
  # including \\*[]
  
  {
    my $str = $tex_str;
    my $space_measure_non_negative_re = qr/\\\\\\*?\[[^-]/;
    
    # Matches a line which is not an empty comment "%".
    # So matches a non-comment line, or matches a comment line which is not
    # empty.
    my $not_an_empty_comment = qr/|[^%].*|%-*[^-\n].*/;
    
    while ($str =~ /\n$not_an_empty_comment\n$space_measure_non_negative_re/g) {
      my $spacing = $2;
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "prefer blank comment line before \\\\[] spacing");
    }
  }
  
  #----------------------------------------------------------------------------
  # check paired \FloatBarrier around \begin{figure} or \begin{my-plain-float}
  
  {
    my %want_next = (none               => 'FloatBarrier',
                     'FloatBarrier'     => 'begin',
                     begin              => 'end',
                     end                => 'FloatBarrier.end',
                     'FloatBarrier.end' => 'FloatBarrier',
                    );
    my $str = $tex_str_sans_comments;
    my $prev_env;
    my $prev = 'none';
    while ($str =~ m{\\((?<barrier>FloatBarrier)
                     |(?<env>begin|end)\{my-plain-float\}
                     )
                  }xg) {
      ### match: $&
      my $this;
      if (defined $+{'barrier'}) {
        $this = $+{'barrier'};
        if (defined $prev_env) {
          $this .= ".$prev_env";
        }
        undef $prev_env;
      } else {
        $this = $+{'env'};
        $prev_env = $this;
      }
      ### $prev
      ### $this
      unless ($want_next{$prev} eq $this) {
        my $linenum = pos_linenum($str);
        bad($linenum, "this $this prev $prev but wanted $want_next{$prev}");
      }
      $prev = $this;
    }
  }
  
  
  #----------------------------------------------------------------------------
  # check for % generated by FILENAME exists
  
  {
    my $str = $tex_str;
    my $dirname = File::Basename::dirname($tex_filename);
    my %seen;
    while ($str =~ /% generated by (\S+)/img) {
      my $filename = $1;
      $filename =~ s/,$//;
      $seen{$filename} = 1;
      my $filename_here = File::Spec->catfile($dirname,$filename);
      my $filename_up   = File::Spec->catfile($dirname, File::Spec->updir,
                                              $filename);
      if (! -e $filename_here && ! -e $filename_up) {
        my $linenum = pos_linenum($str);
        bad_but_not_fatal($linenum, "generated by $filename: does not exist");
      }
    }
    
    foreach my $basename (sort(File::Slurp::read_dir
                               (File::Spec->catdir ($dirname,'tools')))) {
      next unless $basename =~ /^gen-/;
      my $filename = "tools/$basename";
      if (! $seen{$filename}) {
        print "$filename:1: unused generator program\n";
        # $bad = 1;  # not fatal
      }
    }
    
    print scalar(keys %seen)," generator programs\n";
  }
  
  
  #----------------------------------------------------------------------------
  # \index{} in maths mode can come out in .idx with macros spaced out, for
  # example
  #
  #     \begin{align}
  #     \index{foo@\textit{foo}}
  #     \end{align}
  #     =>
  #     \indexentry{foo@\textit  {foo}}{1}     % in .idx file
  #
  # Check here that a \index{} of the same contents is consistently either
  # always in maths mode or always toplevel.  Maths is recognised as
  # "align", "gathered", and \mbox{\index{...}}.  The latter is a good way
  # to force a toplevel to be in maths, if you would otherwise have a
  # mixture.
  
  {
    my $str = $tex_str_sans_comments;
    my $contained_type = 'toplevel';
    my %begin_end_to_type = (begin => 'in-maths',
                             end   => 'toplevel');
    my %seen;
    while ($str =~ /\\(?<begin>begin|end)\{(align|gather)\}
                  |(?<mbox>\\mbox\{)?\\index(?<index>$braces_re)
                   /xsg) {
      if (defined $+{'begin'}) {
        $contained_type = $begin_end_to_type{$+{'begin'}} or die "oops";
        next;
      }
      my $this_type = $contained_type;
      if ($+{'mbox'}) {
        $this_type = 'in-maths';
      }
      my $index = $+{'index'} // die "oops";
      $index =~ s/^\{//;
      $index =~ s/\}$//;
      $index =~ s/\|[()]$//;
      next unless $index =~ /\\/;
      
      my $linenum = pos_linenum($str);
      if (my $seen = $seen{$index}) {
        my ($seen_type,$seen_linenum) = @$seen;
        if ($seen_type ne $this_type) {
          bad($linenum, "\\index $this_type for: $index");
          bad($seen_linenum, "  but previous $seen_type will be different to makeindex");
        }
      } else {
        $seen{$index} = [$this_type,$linenum];
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # check \mathtoolsset preceded by % comment or blank line,
  # so text doesn't flow to it
  
  {
    my $str = $tex_str;
    while ($str =~ /[^\n]\\mathtoolsset/g) {
      my $prev = $1 // '';
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "mathtoolsset not at start of line");
    }
    while ($str =~ /(.*)\n\\mathtools/g) {
      my $prev = $1;
      my $linenum = pos_linenum($str, $+[0]);
      next if $prev =~ /^\s*$/;
      next if $prev =~ /^%/;
      next if $prev =~ /^\\(end|usepackage|vspace)/;
      bad($linenum, "mathtoolsset prefer preceding comment line so no reflow by Emacs");
    }
  }
  
  #----------------------------------------------------------------------------
  # check ;; doubled semis, usually PGF or GP unnecessary extras
  
  {
    my $str = $tex_str;
    while ($str =~ /;{2,}/g) {
      my $whole = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "multi-semicolons $whole");
    }
  }
  
  #----------------------------------------------------------------------------
  # any == >= etc
  # Exceptions made for PGF style
  #     shorten <=.1em       line ends
  #     shorten >=.1em
  #     >=Latex              arrow types
  #     >=Stealth
  #     >={Stealth}
  # But == can occur in tikzmaths so have to exclude all tikz probably.
  
  {
    my $str = $tex_str_sans_comments_and_tikz;
    while ($str =~ /==
                  |(?<!\{)(?<!shorten\ )[<>]=(?!Stealth)(?!\{)
                  |=>/xg) {
      my $op = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "apparent C or GP etc operator: $op");
    }
  }
  
  #----------------------------------------------------------------------------
  # \end{document} duplicate
  
  {
    my $str = $tex_str_sans_comments;
    my $seen = 0;
    my $prev_pos;
    while ($str =~ /\\end\{document\}/g) {
      if (++$seen > 1) {
        my $linenum = pos_linenum($str,$prev_pos);
        bad($linenum, "\\end{document} duplicate");
      }
      $prev_pos = pos($str);
    }
  }
  
  #----------------------------------------------------------------------------
  # \smash{} outside cases or aligned
  
  {
    my %allow_environments = (cases => 1,
                              aligned => 1,
                              tikzpicture => 1);
    my $str = $tex_str;
    my @within;
  MATCH: while ($str =~ /%.*|\\(begin|end)\{([^}]*)\}|\\phantom\{\\smash|\\smash\b(.*)/g) {
      my $begin_end = $1;
      my $environment = $2;
      my $post_smash = $3;
      if ($begin_end) {
        if ($begin_end eq 'begin') {
          push @within, $environment;
        } else {
          if (@within && $within[-1] eq $environment) {
            pop @within;
          } else {
            my $linenum = pos_linenum($str, $+[0]);
            bad($linenum, "unbalanced begin/end: $environment");
          }
        }
      } elsif (defined $post_smash) {
        unless ($post_smash =~ /ALLOW/) {
          foreach my $within (@within) {
            if ($allow_environments{$within}) {
              next MATCH;
            }
          }
          my $linenum = pos_linenum($str, $+[0]);
          bad($linenum, "\\smash outside cases or aligned");
          # print $post_smash,"\n";
        }
      }
    }
  }
  
  
  #----------------------------------------------------------------------------
  # Long GP-Test lines
  
  # Many currently ...
  # {
  #   my $str = $tex_str;
  #   while ($str =~ /^% GP-Test .{72}/mg) {
  #     my $linenum = pos_linenum($str, $+[0]);
  #     bad($linenum, "long GP-Test line");
  #   }
  # }


  #----------------------------------------------------------------------------
  # Double-quote " other than in comments, and other than \" accent.

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /(\")|\\./g) {
      if ($1) {
        my $linenum = pos_linenum($str, $+[0]);
        bad($linenum, "double-quote \" should be `` and ''");
      }
    }
  }

  #----------------------------------------------------------------------------
  # \\begin{scope} duplicate keys
  # some shift= repeats accumulate, other settings override, maybe

  if (0) {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\begin\{scope\}\s*\[([^]]*)/osg) {
      my $params = $1;
      my $pos = $-[0];
      # print "$params\n";
      my %seen;
      while ($params =~ /\s*([^=,]+)(=([^({,]|\(.*?\)|\{.*?\})*)?/osg) {
        my $pos = $pos + $-[1];
        my $key = $1;
        my $value = $2 // '';
        $key =~ s/^,\s*//;
        # print "  key   $key\n";
        # print "  value $value\n";
        if ($seen{$key}++) {
          my $linenum = pos_linenum($str, $pos);
          bad($linenum, "tikz scope duplicate key: $key");
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # \hspace{} after tikzpicture sometimes needs \hbox{\hspace{}}, it seems

  {
    my $str = $tex_str;
    my $space_measure_non_negative_re = qr/\\\\\[[^-]/;
    while ($str =~ /\\end\{tikzpicture\}\n\\(hspace|kern)/g) {
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "may need \\hbox{\\hspace{}} for post-tikz bias");
    }
  }

  #----------------------------------------------------------------------------
  # no tab characters

  {
    my $str = $tex_str;
    while ($str =~ /\t(.*\t)*/g) {
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "tab character, prefer spaces over tabs");
    }
  }

  #----------------------------------------------------------------------------
  # LocalWords ispell

  {
    my $str = $tex_str;
    tie my %LocalWords, 'Tie::IxHash';  # so report in document order
    while ($str =~ /LocalWords:(.*)/g) {
      my $words = $1;
      my $linenum = pos_linenum($str);
      $words =~ s/^[ \t]+|[ \t]+$//g;  # leading and trailing whitespace
      foreach my $word (split / +/, $words) {
        if (exists $LocalWords{$word}) {
          bad($linenum, "duplicate LocalWords: $word");
        } else {
          $LocalWords{$word} = $linenum;
        }
      }
    }
    print scalar(keys %LocalWords)," LocalWords\n";
    $str = $tex_str_sans_comments;
    $str =~ s<\\[-`'"~]><>g;        # sans accents and optional hyphen
    $str =~ s<\\[Hcv]\{(.)\}><$1>g;  # \c{X}
    $str =~ s<\\([lL])\{\}><$1>g;  # \l{} or \L{}
    foreach my $word (keys %LocalWords) {
      unless ($str =~ /[^[:alpha:]]\Q$word\E[^[:alpha:]]/) {
        bad_but_not_fatal($LocalWords{$word}, "unused LocalWords: $word");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check \mathit contents
  # eg. \mathit{foo(x)} should be \mathit{foo}(x)

  {
    my $str = $tex_str_sans_comments;
    my $linebreak_re = qr{\\linebreak\b(\[[0-4]\])?};
    my $mathit_allowed = qr{^([a-zA-Z0-9, ]
                            |\{,\}        # {,} unspaced comma
                            |\{\}         # {} against kerning
                            |\\[ ]        # \ space
                            |$kern_re\b
                            |$linebreak_re
                            |\\(alpha|delta)\b
                            |\\MyNegativePhantom\{\\/\\/\}
                            |\\MyTightDots[ ]
                            )*$}xo;
    my %index;
    while ($str =~ m/mathit($braces_re)/og) {
      my $mathit = $1;
      $mathit =~ s/^\{//;
      $mathit =~ s/\}$//;
      if ($mathit eq '') {
        my $linenum = pos_linenum($str);
        bad($linenum, "empty \\mathit");
      } elsif ($mathit =~ /mathit/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "mathit within mathit: $mathit");
      } elsif ($mathit !~ $mathit_allowed) {
        my $linenum = pos_linenum($str);
        bad($linenum, "bad chars in \\mathit: $mathit");
      }
    }
  }


  #----------------------------------------------------------------------------
  # \end{align} followed by \index{}

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\end\{align\}.*\n?\\index/g) {

      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "\\end{} followed by \\index{} is bad spacing (put it inside the align)");
    }
  }

  #----------------------------------------------------------------------------
  # check \label{foo-formula} outside "align"

  {
    my $str = $tex_str_sans_comments;
    my $within_align = 0;
    my $count_labels = 0;
    my $align_re = qr/(fl)?align/;
    while ($str =~ m<\\(
                       (?<Begin> begin)
                     |(?<End>    end\{$align_re})
                     |(?<My>     MyEquationTag\\)?
                       label\{(?<Label> [^}]*)
                     )>xgo) {
      ### begin: $+{'Begin'}
      if (defined $+{'Begin'}) {
        ### begin align ...
        $within_align = 1;
      } elsif (defined $+{'End'}) {
        ### end align ...
        $within_align = 0;
      } elsif (defined $+{'My'}) {
        ### MyEquationTag ...
        # \MyEquationTag\label{foo} is ok
      } else {
        my $label = $+{'Label'};
        ### $label
        $count_labels++;
        if (! $within_align
            && $label =~ /formula/
            && $label !~ /^page-/) {
          my $linenum = pos_linenum($str);
          bad($linenum, "formula label outside maths: $label");
        }
      }
    }
    if ($verbose) {
      print "$count_labels labels\n";
    }
  }

  #----------------------------------------------------------------------------
  # check \begin{align} etc ending with a \\ separator

  {
    my $str = $tex_str_sans_comments_and_tikz;

    my $whitespace_or_kern_re = qr/$comment_re|\s*|$kern_re|$hspace_re/o;

    while ($str =~ m/\\\\\
                     (\[[^]]*])?
                     ($whitespace_or_kern_re)*
                     \\end\{(?<Type>align|equation|cases)
                    /xmgo) {
      my $type = $+{'Type'};
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "trailing \\\\ in $type");
    }
  }

  #----------------------------------------------------------------------------
  # any \pagebreak reckoned as a leftover

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\pagebreak(?<option>\[\d+\])?(?<allow>\s+% ALLOW)?/g) {
      next if defined $-{'option'};   # \pagebreak[0] is ok
      next if defined $-{'allow'};    # "... % ALLOW" is ok
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "\\pagebreak left-over");
    }
  }


  #----------------------------------------------------------------------------
  # check for lower x. x want \space
  {
    my $str = $tex_str_sans_comments_and_tikz;

    while ($str =~ /[a-z][)']*[.?!] [^ \t]/g) {
      my $word = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "no-end-sentence should be \\ (backslash space): $word");
    }
  }

  # check for abbreviations want .\space
  {
    my $str = $tex_str_sans_comments_and_tikz;

    while ($str =~ /(ie|i\.e)\.(?!\\)/g) {
      my $word = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "abbreviation without \\space: $word");
    }
  }


  #----------------------------------------------------------------------------
  # check \index contents
  # - text@tex parts same
  # - if any $ or \ then it's tex and must have @

  {
    my $str = $tex_str_sans_comments;
    my %index;
    my $braces = $RE{'balanced'}{-parens=>'{}'};
    while ($str =~ m<\\index($braces)>g) {
      my $start = $-[0];
      my $whole = $1;
      $whole = substr($whole,1,-1);
      ### $whole

      if ($whole =~ /[\$\\]/) {
        my $texism = $&;
        if ($whole !~ /\@/) {
          my $linenum = pos_linenum($str, $start);
          bad($linenum, "\\index with $texism TeX but without \@ separator");
          print "  $whole\n";
        }
      }

      foreach my $part (split /!/, $whole) {
        my @ats = split /\@/, $part;
        if (@ats < 2) { next; }
        if (@ats > 2) {
          my $linenum = pos_linenum($str, $start);
          bad($linenum, "\\index multiple \@s: $whole");
          print "   part: $part\n";
          next;
        }
        my ($plain,$tex) = @ats;
        $tex =~ s/\\penalty\d+//g;         # line breaks
        $tex =~ s/\\['"]//g;               # accents
        $tex =~ s/\|[()]//;                # |( and |) index range
        $tex =~ s/\\hyp(\{\}| )?/-/g;      # hypenat \hyp{}
        $tex =~ s/\\MyTightDots/.../g;
        $tex =~ s/\\omega\b/w/g;           # sort \alpha as "w"
        $tex =~ s/\\alpha\b/a/g;           # sort \alpha as "a"
        $tex =~ s/\\leftrightarrow/<->/g;
        $tex =~ s{\\[td]?frac(\d)(\d)}{$1/$2}g;
        $tex =~ s/$kern_re//go;
        $tex =~ s/\\[a-zA-Z]+|[\${}_]//g;  # other macros and specials
        $plain =~ s/^000-//;         # hack for A-numbers
        $plain =~ s/^0+([1-9])/$1/;  # leading 0s for numeric ordering
        $plain =~ s/[ \n]//g;
        $tex =~ s/[ \n]//g;
        unless ($plain eq $tex) {
          my $linenum = pos_linenum($str, $start);
          bad($linenum, "\\index different plain/tex: $whole");
          print "   part: $part\n";
          print "   crunched plain: $plain\n";
          print "   crunched tex:   $tex\n";
        }
      }
    }
  }

  #----------------------------------------------------------------------------
  # superscript or subscript of paren not brace

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /[_^][][()]/g) {
      my $whole = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "subscript/superscript should be braces");
    }
  }

  #----------------------------------------------------------------------------
  # \sqrt99 etc multiple alnum chars after \sqrt should be \sqrt{..}

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\sqrt[0-9][[:alnum:]]+/g) {
      my $whole = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "\\sqrt of multiple chars should have braces");
    }
  }

  #----------------------------------------------------------------------------
  # question mark ? with whitespace before,
  # sometimes from cut and paste of gp output

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /[ \t\r\n]\?/g) {
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "question mark ? with blank before");
    }
  }

  #----------------------------------------------------------------------------
  # empty \\ line break

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\\\[ \t\r\n]+\\\\/g) {
      my $linenum = pos_linenum($str, $+[0]);
      bad($linenum, "empty consecutive \\\\ line breaks");
    }
  }

  #----------------------------------------------------------------------------
  # OEIS should have \textsc{} when in plain text

  # NO
  # {
  #   my $str = $tex_str_sans_comments;
  #   while ($str =~ /(\\textsc\{|title=|-)?\bOEIS\b/g) {
  #     next if defined $1;
  #     my $linenum = pos_linenum($str, $+[0]);
  #     bad($linenum, "$& without \\textsc");
  #   }
  # }

  #----------------------------------------------------------------------------
  # preferring smallcaps OEIS

  # NO
  # {
  #   my $str = $tex_str_sans_comments;
  #   while ($str =~ /\\textsc\{[^}]*}|\b(OEIS)\b/g) {
  #     my $word = $2 // next;
  #     my $linenum = pos_linenum($str, $-[0]);
  #     bad($linenum, "prefer \\textsc{$word}");
  #   }
  # }

  #----------------------------------------------------------------------------
  # check unbalanced lfloor / rfloor
  #
  # occasionally have unbalanced lparen/rparen when padding

  {
    my $str = str_strip_phantom($tex_str);
    my @stack;
    while ($str =~ m{(?<comment> $comments_re)
                   |
                     \\
                     ( (?<dir>    [rl])
                       (?<type>   ceil|floor|brace)
                     | (?<dotdir> left|right)\.
                     )
                     (?<unclosed> \s*%\s*(unclosed|no\ opening))?
                  }xg) {
      if (defined $+{'comment'}) {
        if ($+{'comment'} =~ /ALLOW-UNCLOSED/) {
          pop @stack;
        }
        next;
      }
      my $whole = $&;
      my $dir = $+{'dir'};
      my $type = $+{'type'};
      my $dotdir = $+{'dotdir'};
      my $unclosed = $+{'unclosed'};
      my $linenum = pos_linenum($str);
      ### $whole
      ### $unclosed
      ### $linenum

      next if defined $unclosed;
      if (defined $dotdir) {
        if ($dotdir eq 'left') {
          push @stack, ['.', $linenum];
        } else {
          if (! @stack) {
            bad($linenum, "$whole without matching left");
            next;
          }
          pop @stack;  # \right. matches anything
        }
      } elsif ($dir eq 'l') {
        push @stack, [$type, $linenum];
      } elsif ($dir eq 'r') {
        if (! @stack) {
          bad($linenum, "$whole without matching left");
          next;
        }
        my $elem = pop @stack;
        my ($opp_type,$opp_linenum) = @$elem;
        unless ($opp_type eq $type || $opp_type eq '.') {
          bad($opp_linenum, "unbalanced types $whole");
          bad($linenum,     "  from l$opp_type here");
        }
      } else {
        die;
      }
    }
    while (@stack) {
      my $elem = pop @stack;
      my ($type,$linenum) = @$elem;
      bad($linenum, "unclosed l$type");
    }
  }

  #----------------------------------------------------------------------------
  # check no commented or disguised GP-Test etc

  {
    my $comment_prefix_re
      = qr{^[ \t]*
           (?:
             [\#%*]+    # # Perl, % TeX, * C continuing
           |//+         # // C++
           |/\*+        # /* C
           |=for[ \t]   # =for Perl POD
           )?           # or nothing
        }mx;
    my $str = $tex_str;
    while ($str =~ /$comment_prefix_re[ \t]*((.*?)GP-[-A-Za-z0-9]+)/mgo) {
      my $whole = $1;
      my $disguise = $2;
      next unless $disguise;
      next if $disguise eq 'The ';
      my $linenum = pos_linenum($str, $-[0]);
      ### $&
      ### $whole
      ### $disguise
      ### $linenum
      bad($linenum, "disguised gp-inline: $whole");
    }
  }

  #----------------------------------------------------------------------------
  # check "{=} -" or "{=} +" which will have the sign as operator

  {
    my $str = $tex_str_sans_comments;
    $str =~ s/\\phantom\{=\}//g;
    while ($str =~ /\{=\}\s*([-+]|\\(pm|mp))/g) {
      my $whole = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "{=} with following operator: $whole");
    }
  }

  #----------------------------------------------------------------------------
  # \textit{S} of long word should have \hyphenation

  {
    my %builtin_hyphenation = (direction => 1);

    my %seen;
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\textit\{(.*?)\}/g) {
      my $textit = $1;
      my $linenum = pos_linenum($str, $-[0]);
      foreach my $word (split /\W+/, $textit) {
        next if $builtin_hyphenation{$word};
        next if $hyphenation{$word};
        next if length($word) < 5;
        next if $seen{$word}++;
        bad($linenum, "textit without \\hyphenation: $word");
      }
    }
  }

  #----------------------------------------------------------------------------
  # long words should have \hyphenation

  if (0)        # too many
    {
      my $str = $tex_str_sans_comments_and_tikz;
      $str =~ s<^.*?\\begin\{document\}><str_leave_newlines($&)>es;
      $str =~ s/\\(mathit|url|hyphenation|label|eqref|noeqref|cite|nocite|bibitem)\{.*?\}//g;
      $str =~ s/\\index\{.*?@(.*?)\}/$1/g;
      $str =~ s<\\(begin|end)\{.*?\}><>g;

      my %builtin_hyphenation = (direction => 1);
      require TeX::Hyphen;
      my $hyp = TeX::Hyphen->new
        (file => '/usr/share/texlive/texmf-dist/tex/generic/hyphen/hyphen.tex');
      my %seen;
      while ($str =~ /(\\)?([[:alpha:]]+)/g) {
        next if $1;
        my $word = $2;
        next if length($word) < 8;
        next if $seen{$word}++;
        my $linenum = pos_linenum($str, $-[0]);
        my @splits = $hyp->hyphenate($word);
        next if @splits >= 2;
        bad_but_not_fatal($linenum, "long word without \\hyphenation: $word");

        $word = lc($word);
        if ($word =~ s/(s|ing|ings)$//) {
          my $end = $1;
          my $visualize = $hyp->visualize($word) . $end;
          if ($visualize =~ /-/) {
            print "    maybe: \\hyphenation{$visualize}\n";
          }
        }
      }
    }

  #----------------------------------------------------------------------------
  # check unused hyphenations

  {
    my $str = $tex_str_sans_comments;
    my %unused = %hyphenation;
    while ($str =~ /\w+/g) {
      my $word = $&;
      if (defined $unused{$word}) {
        delete $unused{$word};
        last unless %unused;
      }
    }
    foreach my $word (sort {$hyphenation{$a} <=> $hyphenation{$b}}
                      keys %unused) {
      my $linenum = $hyphenation{$word};
      bad_but_not_fatal($linenum, "unused \\hyphenation: $word");
    }
  }

  #----------------------------------------------------------------------------
  # prefer \operatorname{Re} and Im rather than \mathit{}

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\mathit\{(Re|Im)\}/ig) {
      my $whole = $&;
      my $inside = $1;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "$inside prefer \\operatorname: $whole");
    }
  }

  #----------------------------------------------------------------------------
  # check \mathit{S} of a single character

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\mathit\{.\}/g) {
      my $mathit = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "mathit of single character: $mathit");
    }
  }

  #----------------------------------------------------------------------------
  # check \dotsc not preceded by , comma

  {
    my $str = $tex_str_sans_comments;
    $str =~ s/\\phantom\{\\dotsc\}//g;
    $str =~ s/\\linebreak(\[\d+\])?//g;
    $str =~ s/\\math[rlc]lap\{(\\dotsc)\}/$1/g;
    while ($str =~ m/\\dotsc/g) {
      # print "dotsc at ",pos($str),"\n";
      my $beg = $-[0];
      my $len = min($beg,256);
      my $preceding = substr($str,$beg-$len,$len);
      my $whole = '';
      while ($preceding =~ m<(\s+
                             |\\(degree
                               |[,;]
                               |hspace\{[^}]*\}
                               |$kern_sans_backslash_re
                               )
                             )$>xo) {
        $whole = $& . $whole;
        $preceding = substr($preceding,0,$-[0]);
      }
      unless ($preceding =~ /(,|[0-9]*\.[0-9]+)$/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "\\dotsc not preceded by comma");    # : $whole
      }
    }
  }

  #----------------------------------------------------------------------------
  # check unused \bibitem{}

  {
    my $str = $tex_str_sans_comments;
    my $count_bibitems = 0;
    tie my %bibitems, 'Tie::IxHash';
    while ($str =~ m<bibitem\{([^}]*)}>g) {
      my $bibitem = $1;
      next if $bibitem =~ /#/;  # macro parameters
      my $linenum = pos_linenum($str);
      if (defined $bibitems{$bibitem}) {
        bad($linenum, "duplicate bibitem: $bibitem");
        bad($bibitems{$bibitem}, "was defined here");
      } else {
        $bibitems{$bibitem} = $linenum;
      }
    }

    # check all \cite are to a defined \bibitem
    my %used_bibitems;
    while ($str =~ m<\\(cite)(\[.*?\])?\{([^}]*)}>g) {
      my $bibitem_list = $3;
      next if $bibitem_list =~ /#/;  # macro parameters
      foreach my $bibitem (split /,/, $bibitem_list) {
        if (! exists $bibitems{$bibitem}) {
          my $linenum = pos_linenum($str);
          bad($linenum, "no such bibitem: $bibitem");
        }
        $used_bibitems{$bibitem} = 1;
      }
    }

    # warn on \bititem not used by some \cite
    foreach my $bibitem (keys %bibitems) {
      if (! $used_bibitems{$bibitem}) {
        my $linenum = $bibitems{$bibitem};
        bad_but_not_fatal($linenum, "unused bibitem: $bibitem");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check unused \label{}

  {
    my $str = $tex_str_sans_comments;
    my $count_labels = 0;
    tie my %labels, 'Tie::IxHash';
    while ($str =~ m<\\label\{(.*?)}>g) {
      my $label = $1;
      next if $label =~ /#/;  # skip labels generated with macro parameters
      my $linenum = pos_linenum($str);
      if (defined $labels{$label}) {
        bad($linenum, "duplicate label: $label");
        bad($labels{$label}, "was defined here");
      } else {
        $labels{$label} = $linenum;
      }
    }

    my %used_labels;
    while ($str =~ m<\b(ref|eqref|noeqref|autoref|autopageref)\*?\{([^}]*)}>g) {
      my $label = $2;
      next if $label =~ /#/;  # macro parameters
      if (! exists $labels{$label}) {
        my $linenum = pos_linenum($str);
        bad($linenum, "no such label: $label");
      }
      $used_labels{$label} = 1;
    }

    # # labels used in comments are allowed
    # while ($tex_str =~ m<(eqref|noeqref|autoref|autopageref)\*?\{([^}]*)}>g) {
    #   my $label = $2;
    #   next if $label =~ /#/;  # macro parameters
    #   $used_labels{$label} = 1;
    # }

    foreach my $label (keys %labels) {
      if (! $used_labels{$label}) {
        my $linenum = $labels{$label};
        bad_but_not_fatal($linenum, "unused label: $label");
      }
    }
  }

  #----------------------------------------------------------------------------
  # check for % GP-DEFINE braces { } with backslashing too

  {
    my %redefinition_ok = ('for' => 1,
                           'my'  => 1);
    my $gp_comment_re = qr{\\\\|/\*};
    my $str = $tex_str;
    my $in_brace = 0;
    while ($str =~ /^% GP-DEFINE(.*)/img) {
      my $gp = $1;
      if ($gp =~ /\{/) { $in_brace = 1; }
      if ($in_brace && $gp =~ /\\\s*$/) {
        my $linenum = pos_linenum($str);
        bad($linenum, "GP-DEFINE backslash in braces: $gp");
      }
      if ($gp =~ /\}/) { $in_brace = 0; }
    }
  }

  #----------------------------------------------------------------------------
  # \pagebreak[0] should be at the end of a paragraph

  {
    my $str = $tex_str;

    while ($str =~ m/\\pagebreak
           (\[[^]]*])?       # optional [0] etc
             [ \t]*
             ${comments_re}
                     [ \t]*
                     (?<after>\S*)
                    /xsg) {
      my $after = $+{'after'};
      next if $after =~ /^\\begin\{/ || $after eq '';
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "\\pagebreak not at end of paragraph");
      print "  after: $after\n";
    }
  }

  #----------------------------------------------------------------------------
  # check [my dot] (nodename)

  # {
  #   my $str = $tex_str_sans_comments;
  #
  #   while ($str =~ m/ dot\b[^]\{]*]\s*\(/g) {
  #     my $type = $+{'type'};
  #     my $linenum = pos_linenum($str, $-[0]);
  #     bad($linenum, "[my dot] (nodename)");
  #   }
  # }

  #----------------------------------------------------------------------------
  # check unbackslashed kern

  {
    my $str = $tex_str;

    while ($str =~ /(?<![\\m])kern[-.0-9]|(?<![\\])mkern[-.0-9]/g) {
      my $kern = $&;
      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "unbackslashed kern: $kern");
    }
  }

  #----------------------------------------------------------------------------
  # check doubled words

  {
    # >= 1 whitespace chars, at most 1 newline
    my $whitespace_max_one_newline_re
      = qr/[ \t\r]+(?:\n[ \t\r]*)?|[ \t\r]*\n[ \t\r]+/;

    my $str = $tex_str_sans_comments;

    # delete \begin{tabular}{c l l l}  and {longtable}
    $str =~ s/\{(tabular|longtable)\}.*//g;

    # word of 2 or more chars, separated by whitespace
    while ($str =~ /\b([a-z]\w*)$whitespace_max_one_newline_re(\1)\b/sgio) {
      my $doublon = $1;
      my $other = $2;
      # case insensitive on first char only, rest exact to be bad
      next unless substr($doublon,1) eq substr($other,1);

      my $linenum = pos_linenum($str, $-[0]);
      bad($linenum, "doubled word: $doublon $other");
    }
  }

  #----------------------------------------------------------------------------
  # check for % GP-Test ending in ;

  while ($tex_str =~ /^% (Test-Pari|GP-Test)\s+.*;\s*$/img) {
    my $gp_test = $1;
    my $linenum = pos_linenum($tex_str, $-[0]);
    bad($linenum, "$gp_test end with ;");
  }

  #----------------------------------------------------------------------------
  # check for indented % GP-Test etc

  while ($tex_str =~ /^[ \t]+% ((Test-Pari|GP-)\S*)/mg) {
    my $gp = $1;
    my $linenum = pos_linenum($tex_str, $-[0]);
    bad($linenum, "indented $gp");
  }

  #----------------------------------------------------------------------------
  # check for \frac123 more than 2 digits

  {
    my $str = $tex_str;

    while ($str =~ /\\[td]?frac\d{3,}/g) {
      my $frac = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "$frac more than 2 digits");
    }
  }

  #----------------------------------------------------------------------------
  # check for \eqref{figure- or theorem- as they should be \autoref
  # check for \autoref{...-formula}      as it should be \eqref

  {
    my $str = $tex_str;
    while ($str =~ /\\autoref\{[^}]*?-formula\}/g) {
      my $all = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "$all should be \\eqref");
    }
  }
  {
    my $str = $tex_str;
    while ($str =~ /\\eqref\{(figure|theorem)-.*?\}/g) {
      my $all = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "$all should be \\autoref");
    }
  }

  #----------------------------------------------------------------------------
  # check for hyphens which should use \hyp{}

  {
    my $str = $tex_str_sans_comments_and_tikz;

    # no label part of \eqref{foo-label-xxx}
    # keep [] part of \cite[text text]{foo-label-xxx}
    $str =~ s/(
                \\(hyphenation
                |bibitem|label|tagform|newtagform
                |(cite|ref|eqref|noeqref|autoref|autopageref)\*?
                |url|href|hypertarget|hyperlink
                |(sub)*section
                |MyDOI
                )
                (\[.*?\])?         # optional [foo] extra
              )
              \{[^}]*\}            # label name, not subject to \hyp{}
             /$1/gx;
    $str =~ s/\\usetagform\{.*?\}//g;   # \usetagform{} name
    $str =~ s/\\newfloat\{.*?\}//g;     # \newfloat{} environ name
    $str =~ s/\\(begin|end)\{.*?\}//g;  # \begin{foo} environ names
    $str =~ s/(\\index\{)[^@}]*@/$1/g;  # \index{text@tex} text part skip
    $str =~ s/(\\(index|makeindex|printindex))\[.*?\]/$1/sg;  # index filenames
    $str = str_strip_MyWebPage($str);
    $str =~ s/pdfkeywords=\{[^}]*\}//g;
    $str =~ s/\\MyFurtherProof\{[^}]*}\{[^}]*}//g;

    while ($str =~ /(\\)?([[:alnum:]]+)-([[:alnum:]]+)/ig) {
      next if defined $1;  # macro with - after is ok
      my $pre = $2;
      my $post = $3;
      my $word = $&;
      my $linenum = pos_linenum($str);
      pos($str) = $-[3];   # to consider $post part and further dash

      # 6 or more digits might by hyphenated
      if ($pre =~ /^[[:digit:]]*$/ && $post =~ /^[[:digit:]]*$/) {
        next unless length($pre) >= 6 || length($post) >= 6;
      }

      # 5 or more letters might by hyphenated
      next unless length($pre) >= 5 || length($post) >= 5;

      bad($linenum, "hyphen should use \\hyp{}: ",$word);
    }
  }

  {
    my $str = $tex_str_sans_comments;
    while ($str =~ /\\hyp\{\}-/g) {
      my $linenum = pos_linenum($str);
      bad($linenum, "\\hyp{} with dash");
    }
  }

  #----------------------------------------------------------------------------
  # check \newcommand{\Foo}{123}; ending semicolon

  {
    my $str = $tex_str_sans_comments;

    while ($str =~ /.*(\\newcommand.*\};)[ \t]*(%.*)?$/mg) {
      my $command = $&;
      my $linenum = pos_linenum($str);
      my $col = $+[1] - $-[0] + 1;
      bad("$linenum:$col", "\\newcommand ending semicolon: $command");
    }
  }

  #----------------------------------------------------------------------------
  # check for capital X. missing X\@.
  #
  # ? ! : get extra space like end-of-sentence characters.
  # Think : is ok left alone.
  # Right paren or quote ) ' skipped.
  # TeX Book page 74 ...

  {
    my $str = $tex_str_sans_comments_and_tikz;

    while ($str =~ /([[:alnum:]]*[A-Z][)']*)[.?!]([ \t][ \t]|\n)/g) {
      my $word = $1;
      my $linenum = pos_linenum($str,$-[2]);
      bad($linenum, "end-of-sentence should be: $word\\\@.");
    }
  }

  #----------------------------------------------------------------------------
  # check for sentence ending ".."

  {
    my $str = $tex_str_sans_comments_and_tikz;
    $str =~ s/(\\index\{)[^}]*(@)/$1$2/g;  # not in text part of \index
    while ($str =~ /[.,;]\./g) {
      my $dots = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "two dots: $dots");
    }
  }

  #----------------------------------------------------------------------------
  # check for ",," anywhere except comments (which can be gp comma stuff)

  {
    my $str = $tex_str_sans_comments;

    while ($str =~ /,\s*,/sg) {
      my $commas = $&;
      my $linenum = pos_linenum($str);
      bad($linenum, "two commas: $commas");
    }
  }

  #----------------------------------------------------------------------------
  # check for xx/yy with / instead of \slash or similar
  # two or more chars on each side
  # Some false positives on tikz /sqrt(2) and similar.

  {
    my $str = $tex_str_sans_comments_and_tikz;
    $str = str_strip_MyWebPage($str);
    $str =~ s/(\\(url|href)\{)[^}]*(\})/$1$3/g; # delete guts of \url or \href
    $str =~ s/(\(\$).*?(\$\))/$1 $2/g;  # delete guts of ($ .. $)

    while ($str =~ m{[a-z][a-z]+/([a-z][a-z]+)}ig) {
      my $word = $&;
      my $linenum = pos_linenum($str);
      pos($str) = $-[1];
      bad($linenum, "/ rather than \\slash or similar: $word");
    }
  }

  #----------------------------------------------------------------------------
  # check for multi-chars in maths mode

  {
    ### multi-char words in maths ...

    my $str = $tex_str;
    $str =~ s/^%[^\n]*/%/mg;  # whole line comments become %
    $str =~ s/([^\\\n](\\\\)*)%[^\n]*/$1/mg;  # non-whole-line comments delete
    # ### $str

    my %allow;
    while ($tex_str =~ /^% MY-CHECK-TEX-ALLOW-MULTICHAR: (.*)/mg) {
      ### $1
      @allow{split /\s+/, $1} = ();  # hash slice
    }

    my %maths_environments = (align       => 1,
                              aligned     => 1,
                              cases       => 1,
                              equation    => 1,
                              displaymath => 1,
                             );

    my @state = 0;  # 0=text, 1=maths, 2=tikz
    while ($str
           =~ m!
              \\(?<BeginEnd>begin|end)\{(?<Environment>[^}]*)\}
              |(?<Text>
                  \\(text(bf|sc)?
                    |math(it|rm|bf|sf|tt|cal)
                    |operatorname\*?
                    |raisebox\{[^}]*\}                  #\raisebox{1pt}{...
                    |index|kbd|var
                    |eqref|noeqref|label|intertext|tag\*?
                    |[hvm]space
                    |rule(\[[^]]*\])?\{[^}]*\}          #\rule[]{}{...
                    |genfrac\{[^}]*\}\{[^}]*\}\{[^}]*\} #\genfrac{}{}{}{...
                    |setlength\\[a-zA-Z]+               #\setlength\foo{...
                    )
                    \s*\{
                )
              |(?<Kern>
                  $kern_re                        # \kern.5em, \mkern1mu
                |\\\\\*?\[$dimension_re\]         # \\[.5ex]
                |\\((thin|med|thick)muskip=|abovedisplayskip=)
                  $dimension_re                 # \thickmuskip=1mu
                  )
              |(?<LineBreak>\\\\\*?(\[[a-z0-9.]+\])?)
              |(?<Backslash>\\)?         # possible macro marker
               [[:alpha:]][[:alnum:]]+   # text word, including X123 digits
              |(?<Underscore>[_^]
                  [^^_{}) \t\r\n\\]      # _\macro ok
                  [^^_{}()\[\] \t\r\n,\$=/.+-]+)
              |(?<Star>\*)
              |[\${}]
              |\n\n
              |\\[\$*]       # \$ not maths, \* not a star
              |\(\$          # tikz coordinate open
              |\$\)          # tikz coordinate close
               !igx) {
      my $match = $&;
      my $begin = $+{'BeginEnd'};
      my $environment = $+{'Environment'};
      my $underscore = $+{'Underscore'};
      my $star = $+{'Star'};
      my $kern = $+{'Kern'};
      my $text = $+{'Text'};
      my $backslash = $+{'Backslash'};
      my $linebreak = $+{'LineBreak'};

      ### $match
      if (defined $begin) {
        if ($begin eq 'begin') {
          ### begin environment group: $maths_environments{$environment}
          push @state, $maths_environments{$environment} ? 1 : 0;
        } else {
          ### end group ...
          pop @state;
        }
      } elsif (defined $text) {
        ### text mode open ...
        push @state, 0;

      } elsif (defined $kern) {
        ### kern etc: $kern
      } elsif (defined $backslash) {
        ### other backslash macro ...
      } elsif (defined $linebreak) {
        ### line break \\ macro ...

      } elsif ($match eq '$') {
        ### flip text/maths mode ...
        $state[-1] ^= 1;
      } elsif ($match eq "\n\n") {
        ### blank line, paragraph end ...
        if ($state[-1]) {
          ### oops ...
        }
        $state[-1] = 0;  # text mode at paragraph end


      } elsif ($match eq '{') {
        ### open group ...
        push @state, $state[-1];
      } elsif ($match eq '}') {
        ### close group ...
        pop @state;

      } elsif ($match eq '($') {
        ### tikz calc ...
        push @state, 2;

      } elsif (defined $underscore) {
        if ($state[-1] == 1) {
          my $linenum = pos_linenum($str);
          bad($linenum, "multi-char subscript or superscript: $match");
        }

      } elsif (defined $star) {
        if ($state[-1] == 1) {
          my $linenum = pos_linenum($str);
          bad($linenum, "star * in maths should be \\star usually");
        }

      } elsif ($state[-1]==1 && length($match) > 1) {
        unless (exists $allow{$match}) {
          my $linenum = pos_linenum($str);
          bad($linenum, "multi-char word in maths mode: $match");
        }
      }
    }
  }
}

#------------------------------------------------------------------------------
sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

sub count_lines {
  my ($str) = @_;
  my $count = ($str =~ tr/\n/\n/);
  if ($str =~ /[^\n]$/) { $count++; }
  return $count;
}

# return $str with all characters except \n newlines deleted
sub str_leave_newlines {
  my ($str) = @_;
  $str =~ tr/\n//cd;
  return $str;
}

sub str_strip_MyWebPage {
  my ($str) = @_;
  $str =~ s/\\newcommand\{?\\MyWebPage[^\n]*//;
  return $str;
}

# FIXME: only goes to one nesting level of braces
sub str_strip_phantom {
  my ($str) = @_;
  $str =~ s<\\v?phantom\{[^}]*(\{[^}]*\}[^}]*)*\}>
           <str_leave_newlines($&)>eg;
  return $str;
}

#------------------------------------------------------------------------------

{
  my $help = sub {
    print "$progname [--options] filename...\n";
    my @opts =
      (['-h, --help',    'Print this help'],
       ['-v, --version', 'Print program version'],
       ['--verbose',     'Print extra messages'],
      );
    my $width = 2 + max (map { length ($_->[0]) } @opts);
    foreach (@opts) {
      printf "%-*s%s\n", $width, $_->[0], $_->[1];
    }
    print "\n";
    exit 0;
  };

  GetOptions ('help|?' => $help,
              version => sub {
                print "$progname version $VERSION\n";
                exit 0;
              },
              verbose    => \$verbose,
             )
    or exit 1;

  @ARGV or $help->();
}

while (@ARGV) {
  $tex_filename = shift @ARGV;
  check_file();
}

exit ($bad ? 1 : 0);
